import os
import sys
sys.path.append('')
from dotenv import load_dotenv
import pandas as pd
from hopp.simulation.technologies.sites import SiteInfo
from hopp.simulation.technologies.sites import flatirons_site as sample_site
from hopp.utilities.keys import set_developer_nrel_gov_key
# from plot_reopt_results import plot_reopt_results
# from run_reopt import run_reopt
from greenheart.to_organize.H2_Analysis.hopp_for_h2 import hopp_for_h2
from greenheart.to_organize.H2_Analysis.hopp_for_h2 import run_h2a as run_h2a #no h2a function
from greenheart.to_organize.H2_Analysis.simple_dispatch import SimpleDispatch
from greenheart.to_organize.H2_Analysis.simple_cash_annuals import simple_cash_annuals
import hopp.simulation.technologies.hydrogen.electrolysis.run_h2_PEM as run_h2_PEM
import numpy as np
import numpy_financial as npf
from lcoe.lcoe import lcoe as lcoe_calc
import matplotlib.pyplot as plt
import warnings
warnings.filterwarnings("ignore")

import copy

"""
Perform a LCOH analysis for an offshore wind + Hydrogen PEM system

1. Offshore wind site locations and cost details (4 sites, $1300/kw capex + BOS cost which will come from Orbit Runs)~
2. Cost Scaling Based on Year (Have Weiser et. al report with cost scaling for fixed and floating tech, will implement)
3. Cost Scaling Based on Plant Size (Shields et. Al report)
4. Future Model Development Required:
- Floating Electrolyzer Platform (per turbine vs. centralized)
"""


import os
from pytest import approx, fixture
import numpy as np

from greenheart.simulation.technologies.hydrogen.electrolysis.PEM_BOP.PEM_BOP import (
    pem_bop,
)
from greenheart.simulation.greenheart_simulation import (
    run_simulation,
    GreenHeartSimulationConfig,
)

from ORBIT.core.library import initialize_library

@fixture
def bop_energy():
    power_profile_kw = np.array(
        [
            0,
            9999,  # just below turndown ratio
            10000,  # exactly at turndown ratio
            82746,
            93774,  # max power in csv. largest operating ratio
            100000,  # full power
        ]
    )

    electrolyzer_rating_mw = 100  # MW
    turndown_ratio = 0.1

    bop_energy = pem_bop(
        power_profile_to_electrolyzer_kw=power_profile_kw,
        electrolyzer_rated_mw=electrolyzer_rating_mw,
        electrolyzer_turn_down_ratio=turndown_ratio,
    )
    return bop_energy


def test_bop_energy(subtests, bop_energy):
    with subtests.test("No power"):
        assert bop_energy[0] == 0
    with subtests.test("below turndown"):
        assert bop_energy[1] == 0
    with subtests.test("at turndown"):
        assert bop_energy[2] == approx(11032.668, 1e-2)
    with subtests.test("mid-range power"):
        assert bop_energy[3] == approx(6866.87, 1e-2)
    with subtests.test("max power in curve"):
        assert bop_energy[4] == approx(7847.85)
    with subtests.test("full power"):
        assert bop_energy[5] == approx(7847.85)

dirname = os.path.dirname(__file__)
parent_dir = os.path.dirname(dirname)

library_path = os.path.join(parent_dir, "input_files/")

initialize_library(library_path)

turbine_model = "osw_18MW"
filename_turbine_config = os.path.join(
    library_path, f"turbines/{turbine_model}.yaml"
)
filename_orbit_config = os.path.join(
    library_path, f"plant/orbit-config-{turbine_model}-stripped.yaml"
)
filename_floris_config = os.path.join(
    library_path, f"floris/floris_input_{turbine_model}.yaml"
)
filename_greenheart_config = os.path.join(
    library_path, f"plant/greenheart_config.yaml"
)

filename_hopp_config = os.path.join(
    library_path, f"plant/hopp_config.yaml"
)

def test_greenheart_simulation_pem_bop(subtests):
    config = GreenHeartSimulationConfig(
        filename_hopp_config=filename_hopp_config,
        filename_greenheart_config=filename_greenheart_config,
        filename_turbine_config=filename_turbine_config,
        filename_orbit_config=filename_orbit_config,
        filename_floris_config=filename_floris_config,
        verbose=False,
        show_plots=False,
        save_plots=False,
        use_profast=True,
        post_processing=True,
        incentive_option=1,
        plant_design_scenario=1,
        output_level=3,
    )
    lcoh,_,_,_,_,_,_,annual_energy_breakdown = run_simulation(config)

    # include electrolyzer bop power consumption in greenheart simulation
    config.greenheart_config['electrolyzer']["include_bop_power"] = True

    lcoh2,_,_,_,_,_,_,annual_energy_breakdown2 = run_simulation(config)

    with subtests.test("LCOH not equal"):
        assert lcoh != lcoh2

    with subtests.test("annual_energy_breakdown no electrolyzer bop"):
        assert annual_energy_breakdown["electrolyzer_bop_energy_kwh"] == 0

    with subtests.test("annual_energy_breakdown electrolyzer bop"):
        assert annual_energy_breakdown2["electrolyzer_bop_energy_kwh"] == approx(79921217.81100339,1e-3)


import os
import pathlib
from hopp.simulation.hybrid_simulation import HybridSimulation, TechnologiesConfig
import json
from hopp.tools.analysis import create_cost_calculator
from hopp.simulation.hopp_interface import HoppInterface
import pandas as pd

def hopp_for_h2_floris(site, scenario, technologies, wind_size_mw, solar_size_mw, storage_size_mw, storage_size_mwh, storage_hours,
                wind_cost_kw, solar_cost_kw, storage_cost_kw, storage_cost_kwh,
                kw_continuous, load,
                custom_powercurve,
                interconnection_size_mw, grid_connected_hopp=True, wind_om_cost_kw=42, solar_om_cost_kw=0, turbine_parent_path=None, ppa_price=0.0,
                wave_cost_dict={}):
    '''
    Runs HOPP for H2 analysis purposes
    :param site: :class:`hybrid.sites.site_info.SiteInfo`,
        Hybrid plant site information which includes layout, location and resource data
    :param scenario: ``dict``,
        Dictionary of scenario options, includes location, year, technology pricing
    :param technologies: nested ``dict``; i.e., ``{'pv': {'system_capacity_kw': float}}``
        Names of technologies to include and configuration dictionaries
        For details on configurations dictionaries see:
            ===============   =============================================
            Technology key    Class for reference
            ===============   =============================================
            ``pv``            :class:`hybrid.pv_source.PVPlant`
            ``wind``          :class:`hybrid.wind_source.WindPlant`
            ``tower``         :class:`hybrid.tower_source.TowerPlant`
            ``trough``        :class:`hybrid.trough_source.TroughPlant`
            ``battery``       :class:`hybrid.battery.Battery`
            ===============   =============================================
    :param wind_size_mw: ``float``,
        Wind technology size in MW
    :param solar_size_mw: ``float``,
        Solar technology size in MW
    :param storage_size_mw: ``float``,
        Storage technology size in MW
    :param storage_size_mwh: ``float``,
        Storage technology size in MWh
    :param storage_hours: ``float``,
        Number of hours of storage at maximum output rating.
    :param wind_cost_kw: ``float``,
        Wind installed cost in $/kw
    :param solar_cost_kw: ``float``,
        Solar installed cost in $/kW
    :param storage_cost_kw: ``float``,
        Storage cost in $/kW
    :param storage_cost_kwh: ``float``,
        Storage cost in $/kWh
    :param kw_continuous: ``float``,
        kW rating of electrolyzer
    :param load: ``list``,
        (8760) hourly load profile of electrolyzer in kW. Default is continuous load at kw_continuous rating
    :param custom_powercurve: ``bool``,
        Flag to determine if custom wind turbine powercurve file is loaded
    :param interconnection_size_mw: ``float``,
        Interconnection size in MW
    :param grid_connected_hopp: ``bool``,
        Flag for on-grid operation. Enables buying/selling of energy to grid.
    :returns:

    :param hybrid_plant: :class: `hybrid.hybrid_simulation.HybridSimulation`,
        Base class for simulation a Hybrid Plant
    :param combined_hybrid_power_production_hopp: ``list``,
        (8760x1) hourly sequence of combined pv and wind power in kW
    :param combined_hybrid_curtailment_hopp: ``list``,
        (8760x1) hourly sequence of combined pv and wind curtailment/spilled energy in kW
    :param energy_shortfall_hopp: ``list``,
        (8760x1) hourly sequence of energy shortfall vs. load in kW
    :param annual_energies: ``dict``,
        Dictionary of AEP for each technology
    :param wind_plus_solar_npv: ``float``,
        Combined Net present value of wind + solar technologies
    :param npvs: ``dict``,
        Dictionary of net present values of technologies
    :param lcoe: ``float``
        Levelized cost of electricity for hybrid plant
    '''

    # Create model
    dispatch_options = {'battery_dispatch': 'heuristic'}
    if 'grid' not in technologies.keys():
        technologies['grid'] = {'interconnect_kw': interconnection_size_mw * 1e3}
    
    data = {"technologies": technologies,
            "site": site}
    # print(data)
    # create HOPP instance
    hi = HoppInterface(data)
    # hybrid_plant = HybridSimulation(site, technologies, dispatch_options=dispatch_options)
    hi.system.setup_cost_calculator(create_cost_calculator(interconnection_size_mw,
                                                              bos_cost_source='CostPerMW',
                                                              wind_installed_cost_mw=wind_cost_kw * 1000,
                                                              solar_installed_cost_mw=solar_cost_kw * 1000,
                                                              storage_installed_cost_mw=storage_cost_kw * 1000,
                                                              storage_installed_cost_mwh=storage_cost_kwh * 1000
                                                              ))
    
    if 'wave' in technologies.keys():
        hi.system.wave.create_mhk_cost_calculator(wave_cost_dict)

    hi.system.set_om_costs_per_kw(pv_om_per_kw=solar_om_cost_kw, wind_om_per_kw=wind_om_cost_kw, hybrid_om_per_kw=None)
    if solar_size_mw > 0:
        hi.system.pv._financial_model.FinancialParameters.analysis_period = scenario['Useful Life']
        hi.system.pv._financial_model.FinancialParameters.debt_percent = scenario['Debt Equity']
        # hybrid_plant.pv.system_capacity_kw = solar_size_mw * 1000
        # if scenario['ITC Available']:
        #     hybrid_plant.pv._financial_model.TaxCreditIncentives.itc_fed_percent = 26
        # else:
        #     hybrid_plant.pv._financial_model.TaxCreditIncentives.itc_fed_percent = 0

    # TODO clean this up for better function with custom financial model
    if 'wind' in technologies and 'fin_model' in technologies["grid"].keys() and type(technologies["grid"]["fin_model"]) == {}:
        # hybrid_plant.wind._system_model.Turbine.wind_resource_shear = 0.33
        # hybrid_plant.wind.wake_model = 3
        # hybrid_plant.wind.value("wake_int_loss", 3)
        hi.system.wind._financial_model.FinancialParameters.analysis_period = scenario['Useful Life']
        hi.system.wind._financial_model.FinancialParameters.system_capacity = wind_size_mw * 1000
        # hybrid_plant.wind.om_capacity =
        hi.system.wind._financial_model.FinancialParameters.debt_percent = scenario['Debt Equity']
        hi.system.wind._financial_model.value("debt_option", 0)
        hi.system.wind._financial_model.FinancialParameters.debt_percent = scenario['Debt Equity']
        hi.system.wind._financial_model.value("debt_option", 0)
        ptc_val = scenario['Wind PTC']

        interim_list = list(
            hi.system.wind._financial_model.TaxCreditIncentives.ptc_fed_amount)
        interim_list[0] = ptc_val
        hi.system.wind._financial_model.TaxCreditIncentives.ptc_fed_amount = tuple(interim_list)
        #hybrid_plant.wind._system_model.Turbine.wind_turbine_hub_ht = scenario['Tower Height']

        hi.system.wind._financial_model.TaxCreditIncentives.itc_fed_percent = scenario['Wind ITC']
        hi.system.wind._financial_model.FinancialParameters.real_discount_rate = 7
    if custom_powercurve:
        if turbine_parent_path == None:
            turbine_parent_path = os.path.abspath(os.path.dirname(__file__))
        powercurve_file = open(os.path.join(turbine_parent_path, scenario['Powercurve File']))
        powercurve_file_extension = pathlib.Path(os.path.join(turbine_parent_path, scenario['Powercurve File'])).suffix
        if powercurve_file_extension == '.csv':
            curve_data = pd.read_csv(os.path.join(turbine_parent_path, scenario['Powercurve File']))
            wind_speed = curve_data['Wind Speed [m/s]'].values.tolist()
            curve_power = curve_data['Power [kW]']
            hi.system.wind._system_model.Turbine.wind_turbine_powercurve_windspeeds = wind_speed
            hi.system.wind._system_model.Turbine.wind_turbine_powercurve_powerout = curve_power

        else:
            powercurve_data = json.load(powercurve_file)
            powercurve_file.close()
            hi.system.wind._system_model.Turbine.wind_turbine_powercurve_windspeeds = \
                powercurve_data['turbine_powercurve_specification']['wind_speed_ms']
            hi.system.wind._system_model.Turbine.wind_turbine_powercurve_powerout = \
                powercurve_data['turbine_powercurve_specification']['turbine_power_output']

    if 'wind' in technologies: # this was a contested line in a refactor merge - may cause issue
        hi.system.wind.system_capacity_by_num_turbines(wind_size_mw * 1000)
    hi.system.ppa_price = ppa_price
    hi.system.simulate(scenario['Useful Life'])

    # HOPP Specific Energy Metrics
    ### wind wind losses, for wind only farms ###
    combined_hybrid_power_production_hopp = hi.system.grid._system_model.Outputs.system_pre_interconnect_kwac[0:8759]
    energy_shortfall_hopp = [x - y for x, y in
                             zip(load,combined_hybrid_power_production_hopp)]
    energy_shortfall_hopp = [x if x > 0 else 0 for x in energy_shortfall_hopp]
    combined_hybrid_curtailment_hopp = [x - y for x, y in
                             zip(combined_hybrid_power_production_hopp,load)]
    combined_hybrid_curtailment_hopp = [x if x > 0 else 0 for x in combined_hybrid_curtailment_hopp]

    # super simple dispatch battery model with no forecasting TODO: add forecasting
    # print("Length of 'energy_shortfall_hopp is {}".format(len(energy_shortfall_hopp)))
    # print("Length of 'combined_hybrid_curtailment_hopp is {}".format(len(combined_hybrid_curtailment_hopp)))
    # TODO: Fix bug in dispatch model that errors when first curtailment >0
    combined_hybrid_curtailment_hopp[0] = 0

    # Save the outputs
    annual_energies = hi.system.annual_energies
    wind_plus_solar_npv = hi.system.net_present_values.hybrid
    npvs = hi.system.net_present_values
    lcoe = hi.system.lcoe_real
    lcoe_nom = hi.system.lcoe_nom
    # print('lcoe nominal: ', lcoe_nom)
    # print('annual energy',annual_energies)
    # print('discount rate', hybrid_plant.wind._financial_model.FinancialParameters.real_discount_rate)

    return hi.system, combined_hybrid_power_production_hopp, combined_hybrid_curtailment_hopp, \
           energy_shortfall_hopp,\
           annual_energies, wind_plus_solar_npv, npvs, lcoe, lcoe_nom


from greenheart.simulation.technologies.offshore.fixed_platform import FixedPlatformDesign, FixedPlatformInstallation
from greenheart.simulation.technologies.offshore.floating_platform import FloatingPlatformDesign, FloatingPlatformInstallation



import numpy as np

def simple_cash_annuals(plant_useful_life, equipment_useful_life, capex, opex, amortization_interest):
   
    # Number of cycles for equipment
    full_cycle = plant_useful_life // equipment_useful_life
    partial_cycle = plant_useful_life % equipment_useful_life
    
    #Payment period for equipment (assume payment period is life of equipment)
    a = amortization_interest
    amortization = []
    
    for i in range(full_cycle*equipment_useful_life):   
        amortization.append(capex * ((a*(1+a)**equipment_useful_life)\
            /(((1+a)**equipment_useful_life)-1)))
    
    #Payment period if plant's useful life is sorter than equipment 
    for i in range(full_cycle*equipment_useful_life, plant_useful_life):
        amortization.append(capex * ((a*(1+a)**(partial_cycle))\
            /(((1+a)**(partial_cycle))-1)))

    # Initialize annual cash flow array for useful life of plant
    opex_annuals = [opex] * plant_useful_life

    cash_flow_annuals = np.add(opex_annuals, amortization)
    # print(cash_flow_annuals)
    return cash_flow_annuals


if __name__ == '__main__': 
    simple_cash_annuals(30,8, 1000,20,0.03)


from pyomo.environ import *
import random
import numpy as np
import matplotlib.pyplot as plt
import time
import os


def optimize(
    P_wind_t,
    T=50,
    n_stacks=3,
    c_wp=0,
    c_sw=12,
    rated_power=500,
    dt=1,
    P_init=None,
    I_init=None,
    T_init=None,
    AC_init=0,
    F_tot_init=0,
):
    model = ConcreteModel()
    # Things to solve:
    C_INV = 1.47e6
    LT = 90000

    # Initializations
    if P_init is None:
        P_init = [0 for i in range(n_stacks * T)]
    else:
        P_init = P_init.flatten()
    if I_init is None:
        I_init = [0 for i in range(n_stacks * T)]
    else:
        I_init = I_init.flatten()
    if T_init is None:
        T_init = [0 for i in range(n_stacks * T)]
    else:
        T_init = T_init.flatten()

    model.p = Var(
        [i for i in range(n_stacks * T)],
        bounds=(-1e-2, rated_power),
        initialize=P_init,
    )
    model.I = Var(
        [i for i in range(n_stacks * T)],
        within=Binary,
        initialize=I_init,  # .astype(int),
    )
    model.T = Var(
        [i for i in range(n_stacks * T)],
        within=Binary,
        initialize=T_init,  # .astype(int),
    )
    model.AC = Var(
        [0], bounds=(1e-3, 1.2 * rated_power * n_stacks * T), initialize=float(AC_init)
    )
    model.F_tot = Var(
        [0], bounds=(1e-3, 8 * rated_power * n_stacks * T), initialize=float(F_tot_init)
    )
    model.eps = Param(initialize=1, mutable=True)

    C_WP = c_wp * np.ones(
        T,
    )  # could vary with time
    C_SW = c_sw * np.ones(
        T,
    )  # could vary with time

    P_max = rated_power
    P_min = 0.1 * rated_power

    def obj(model):
        return model.AC[0] - model.eps * model.F_tot[0]

    def physical_constraint_AC(model):
        AC = 0
        for t in range(T):
            for stack in range(n_stacks):
                AC = (
                    AC
                    + C_WP[t] * model.p[t * n_stacks + stack]
                    + C_SW[t] * model.T[t * n_stacks + stack]
                )
        return model.AC[0] == AC + C_INV * n_stacks / LT

    def physical_constraint_F_tot(model):
        """Objective function"""
        F_tot = 0
        for t in range(T):
            for stack in range(n_stacks):
                F_tot = (
                    F_tot
                    + (
                        0.0145 * model.p[t * n_stacks + stack]
                        + 0.3874 * model.I[t * n_stacks + stack] * rated_power / 500
                    )
                    * dt
                )
        return model.F_tot[0] == F_tot

    def power_constraint(model, t):
        """Make sure sum of stack powers is below available wind power."""
        power_full_stack = 0
        for stack in range(n_stacks):
            power_full_stack = power_full_stack + model.p[t * n_stacks + stack]
        return power_full_stack <= P_wind_t[t]

    def safety_bounds_lower(model, t, stack):
        """Make sure input powers don't exceed safety bounds."""
        return P_min * model.I[t * n_stacks + stack] <= model.p[t * n_stacks + stack]

    def safety_bounds_upper(model, t, stack):
        """Make sure input powers don't exceed safety bounds."""
        return P_max * model.I[t * n_stacks + stack] >= model.p[t * n_stacks + stack]

    def switching_constraint_pos(model, stack, t):
        trans = model.I[t * n_stacks + stack] - model.I[(t - 1) * n_stacks + stack]
        return model.T[t * n_stacks + stack] >= trans

    def switching_constraint_neg(model, stack, t):
        trans = model.I[t * n_stacks + stack] - model.I[(t - 1) * n_stacks + stack]
        return -model.T[t * n_stacks + stack] <= trans

    model.pwr_constraints = ConstraintList()
    model.safety_constraints = ConstraintList()
    model.switching_constraints = ConstraintList()
    model.physical_constraints = ConstraintList()

    for t in range(T):
        model.pwr_constraints.add(power_constraint(model, t))
        for stack in range(n_stacks):
            model.safety_constraints.add(safety_bounds_lower(model, t, stack))
            model.safety_constraints.add(safety_bounds_upper(model, t, stack))

            if t > 0:
                model.switching_constraints.add(
                    switching_constraint_pos(model, stack, t)
                )
                model.switching_constraints.add(
                    switching_constraint_neg(model, stack, t)
                )
    model.physical_constraints.add(physical_constraint_F_tot(model))
    model.physical_constraints.add(physical_constraint_AC(model))
    model.objective = Objective(expr=obj(model), sense=minimize)
    eps = 10
    cbc_path = os.path.abspath('')+'\\hybrid\\PEM_Model_2Push\\cbc.exe'
    solver = SolverFactory('cbc', executable=cbc_path) # Use this if you have a Windows machine; also make sure that cbc.exe is in the same folder as this script
    #solver = SolverFactory("cbc") # Use this if you don't have a windows machine
    j = 1
    while eps > 1e-3:
        start = time.process_time()
        results = solver.solve(model)
        print("time to solve", time.process_time() - start)

        model.eps = value(model.AC[0] / model.F_tot[0])  # optimal value
        eps = model.AC[0].value - model.eps.value * model.F_tot[0].value
        j = j + 1

    I = np.array([model.I[i].value for i in range(n_stacks * T)])
    I_ = np.reshape(I, (T, n_stacks))
    P = np.array([model.p[i].value for i in range(n_stacks * T)])
    P_ = np.reshape(P, (T, n_stacks))
    Tr = np.array([model.T[i].value for i in range(n_stacks * T)]).reshape(
        (T, n_stacks)
    )
    P_tot_opt = np.sum(P_, axis=1)
    H2f = np.zeros((T, n_stacks))
    for stack in range(n_stacks):
        H2f[:, stack] = (
            0.0145 * P_[:, stack] + 0.3874 * I_[:, stack] * rated_power / 500
        ) * dt
    return (
        P_tot_opt,
        P_,
        H2f,
        I_,
        Tr,
        P_wind_t,
        model.AC[0].value,
        model.F_tot[0].value,
    )
